/*
 * cprw - computed power
 * a "system indipendent" way to estimate your pure computational power
 * by Wave++
 *
 * notes:
 * - we use macros, we must avoid any tipe of optimization that the
 *   compiler might use
 * - if you run it in a multitasking environment (like unix) run it as root
 *   and use nice -10 to set it to realtime priority.
 *
 * what it does:
 * - tests on bit shifting using: short, int, long, long long
 * - tests on bitwise operatos: short, int, long, long long
 * - tests on standard math operators on: short, int, long, long long, float
 *   double, long double
 * - massive allocation tests (ram DOES matter)
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

/* tests defaults */
#define STD_NOPS    10000  /* number of operation for each element */
#define STD_ARRSIZE 500000 /* number of elements per test */
#define STD_MLOOPS  10     /* number of loops for accuracy */

#define BITSHIFT(mType, nOps, arrSize) \
{ \
  mType* buf = calloc(arrSize, sizeof(mType)); \
  unsigned long cOps, bSkew; \
  for(cOps = 0; cOps < nOps; cOps++) \
    for (bSkew = 0; bSkew < sizeof(mType); bSkew++) \
    { \
      buf[cOps] >>= bSkew; \
      buf[arrSize-1-cOps] <<= bSkew; \
    } \
  free(buf); \
}

#define BITWISE(mType, nOps, arrSize) \
{ \
  mType* buf = calloc(arrSize, sizeof(mType)); \
  unsigned long cOps, bSkew; \
  for(cOps = 0; cOps < nOps; cOps++) \
    for (bSkew = 0; bSkew < nOps; bSkew++) \
    { \
      buf[cOps] |= bSkew; \
      buf[cOps] &= bSkew; \
      buf[cOps] ^= bSkew; \
    } \
  free(buf); \
}

#define MATHOPS(mType, nOps, arrSize) \
{ \
  mType* buf = calloc(arrSize, sizeof(mType)); \
  unsigned long cOps, bSkew; \
  for(cOps = 0; cOps < nOps; cOps++) \
    for (bSkew = 0; bSkew < nOps; bSkew++) \
    { \
      buf[cOps] += bSkew; \
      buf[arrSize-1-cOps] -= bSkew; \
      buf[cOps] *= bSkew; \
      buf[arrSize-1-cOps] /= ++bSkew; \
    } \
  free(buf); \
}

#define MALLOCTESTS(mType, nOps, arrSize) \
{ \
  mType* buf; \
  unsigned long cOps, bSkew; \
  for(cOps = 0; cOps < nOps; cOps++) \
    for (bSkew = 1; bSkew < nOps; bSkew*=2) \
    { \
      buf = calloc(bSkew, sizeof(mType)); \
      free(buf); \
    } \
}

#define RUNINTTESTS(testType, nOps, arrSize) \
{ \
  testType(short, nOps, arrSize); \
  testType(int, nOps, arrSize); \
  testType(long, nOps, arrSize); \
  testType(long long, nOps, arrSize); \
}

#define RUNFULLTESTS(testType, nOps, arrSize) \
{ \
  testType(short, nOps, arrSize); \
  testType(int, nOps, arrSize); \
  testType(long, nOps, arrSize); \
  testType(long long, nOps, arrSize); \
  testType(float, nOps, arrSize); \
  testType(double, nOps, arrSize); \
  testType(long double, nOps, arrSize); \
}

#define RUNALL() \
{ \
  RUNINTTESTS(BITSHIFT, STD_NOPS, STD_ARRSIZE); \
  RUNINTTESTS(BITWISE, STD_NOPS, STD_ARRSIZE); \
  RUNFULLTESTS(MATHOPS, STD_NOPS, STD_ARRSIZE); \
  RUNFULLTESTS(MALLOCTESTS, STD_NOPS, STD_ARRSIZE); \
}

unsigned long long timeDiff(const struct timeval *a, const struct timeval *b)
{
  unsigned long long timeA = (unsigned long long)a->tv_sec *
    (unsigned long long)1000000 + (unsigned long long)a->tv_usec;
  unsigned long long timeB = (unsigned long long)b->tv_sec *
    (unsigned long long)1000000 + (unsigned long long)b->tv_usec;

  return timeB-timeA;
}

int main()
{
  unsigned long tLoops;
  struct timeval a, b;

  printf("cpwr estimating...\n");
  
  gettimeofday(&a, NULL);
  for (tLoops = 0; tLoops < STD_MLOOPS; tLoops++)
    RUNALL();
  gettimeofday(&b, NULL);

  printf("time consumption: %llu\n", (timeDiff(&a,&b)/STD_MLOOPS));

  return 0;
}
